#include <extdll.h>
#include <sdk_util.h>
#include <meta_api.h>

#include "plugin.h"

edict_t * pSearchEntity = 0;

static edict_t * FindEntityInSphere_Post(edict_t *pEdictStartSearchAfter, const float *org, float rad) {
	pSearchEntity = (META_RESULT_STATUS == MRES_OVERRIDE) ? META_RESULT_OVERRIDE_RET(edict_t*) : META_RESULT_ORIG_RET(edict_t*);
	
	RETURN_META_VALUE(MRES_HANDLED, 0);
}

static void TraceLine(const float *__v1, const float *__v2, int fNoMonsters, edict_t *pentToSkip, TraceResult *trace) {
	if(FNullEnt(pSearchEntity))
		RETURN_META(MRES_IGNORED);
	
	(*g_engfuncs.pfnTraceLine)(__v1, __v2, fNoMonsters, pentToSkip, trace);
	
	const char * logformat = "Ghost mine explosion detected and blocked. (bug type: %i, %s bug)";
	
	//
	// TraceLine inside "FindEntityIsSphere"-search
	//   This only happens when combat.cpp:RadiusDamage is called.
	//   GhostMine bug happens because RadiusDamage assumes that, if trace->fStartSolid is set 1,
	//   traceline started inside target entity (pSearchEntity). Problem is that trace->fStartSolid 
	//   can been set 1, even if traceline hasn't stopped at this position.
	//
	
	// Check TraceLine result
	if((trace->flFraction == 1.0 || trace->pHit == pSearchEntity) && trace->fStartSolid) {
		if(!trace->fAllSolid || trace->fInOpen) {
			//Bug number 1. Bug in gamedll: Should check fAllSolid instead of fStartSolid.
			trace->fStartSolid = 0;
			
			LOG_MESSAGE(PLID, logformat, 1, "gamedll");
		} 
		else if(trace->fAllSolid && trace->flFraction == 1.0) {
			//Bug number 2. Bug in engine: TraceLine starts from solid (not 'fAllSolid') and hits this
			// same solid (bsp object). fAllSolid is set to '1', and flFraction is set to '1.0' (this is
			// the bug).
			trace->flFraction = 0.0;
			
			LOG_MESSAGE(PLID, logformat, 2, "engine");
		}
	}
	
	// Only need fix for the very first "TraceLine" after "FindEntityIsSphere"
	pSearchEntity = 0;
	
	RETURN_META(MRES_SUPERCEDE);
}

int plugin_GetEngineFunctions(enginefuncs_t *pengfuncsFromEngine, int *interfaceVersion) {
	memset(pengfuncsFromEngine, 0, sizeof(enginefuncs_t));
	
	pengfuncsFromEngine->pfnTraceLine = TraceLine;
	
	return(TRUE);
}

int plugin_GetEngineFunctions_Post(enginefuncs_t *pengfuncsFromEngine, int *interfaceVersion) {
	memset(pengfuncsFromEngine, 0, sizeof(enginefuncs_t));
	
	pengfuncsFromEngine->pfnFindEntityInSphere = FindEntityInSphere_Post;
	
	return(TRUE);
}
